/*
Copyright (c) 2000,2001, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file SenseHandler.C
<pre>
<b>File:</b>          SenseHandler.C
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn 2001
<b>Authors:</b>       Jelle Kok and Remco de Boer
<b>Created:</b>       28/11/2000
<b>Last Revision:</b> 06/01/2002
<b>Contents:</b>      This file contains the class SenseHandler that is used
               to process the information coming from the server.
</pre>
*/

#include "SenseHandler.h"
#include "ActHandler.h"  // sigalarmHandler
#include "Parse.h"

#include <signal.h>   // needed for SIGALARM
#include <string.h>   // needed for strlen
#include <stdio.h>    // needed for printf
#include <iostream.h> // needed for cout

/******************************************************************************/
/********************** CLASS SENSEHANDLER ************************************/
/******************************************************************************/

/*! This function is needed to start the Sense Thread (thread that continually
    waits for input and parses the input). This function is needed since it is
    not possible to call a method from a class using a thread. So this function
    calls handleMessagesFromServer from the SenseHandler class.
    \param v pointer to a SenseHandler class.*/
void* sense_callback( void *v )
{
  Log.log( 1, "Starting to listen for server messages" );
  SenseHandler* s = (SenseHandler*)v;
  s->handleMessagesFromServer( );
  return NULL;
}


/*! Constructor for the SenseHandler. It needs a reference to a connection and
    a reference to a worldmodel.
    \param c Connection from which input is received
    \param wm WorldModel to which new information will be sent for processing
    \param ss ServerSettings that contain the parameters used by the server
    \param ps PlayerSettings that determine how to interact with messages. */
SenseHandler::SenseHandler( Connection *c, WorldModel *wm, ServerSettings *ss,
                            PlayerSettings *ps )
{
  connection             = c;
  SS                     = ss;
  PS                     = ps;
  WM                     = wm;
  iSimStep               = SS->getSimulatorStep()*1000;
  iTimeSignal            = (int)(iSimStep*0.85);

  struct sigaction sigact;

  sigact.sa_flags = SA_RESTART; // primitives (recvfrom) should not be unblocked
  sigact.sa_handler = (void (*)(int))sigalarmHandler;
  sigaction( SIGALRM, &sigact, NULL );

  // set timer signal to indicate when ActHandler should sent commands to the
  // server, this structure will later be filled with exact timing values
  itv.it_interval.tv_sec  = 0;
  itv.it_interval.tv_usec = 0;
  itv.it_value.tv_sec     = 0;
  itv.it_value.tv_usec    = 0;
}

/*! This is the main routine of this class. It loops forever (till the thread
    is destroyed) and receives and parses the incoming messages.   */
void SenseHandler::handleMessagesFromServer( )
{
  char strBuf[MAX_MSG];
  int i=0;

  while( 1 )
  {
    strBuf[0]='\0';
    if( i != -1 )                                         // if no error
      i = connection->receiveMessage( strBuf, MAX_MSG );  // get message
    if( strBuf[0] != '\0' )                               // if not empty
      analyzeMessage( strBuf );                           // parse message
  }
}


/*! This method sets the time signal. This is the time that should be waited
    before the next action should be sent to the server. As soon as a sense
    message arrives this method is called. Using the information from the member
    variable 'iTriCounter' which denotes when the see message will arrive in
    this cycle (0=first half, 1=2nd half, 2=no see, all for the default view
    frequency) the timer is set. The values that denote the fraction of the
    simulation step that is waited are all defined in PlayerSettings, such that
    they can be easily changed. */
void SenseHandler::setTimeSignal( )
{
  if( WM->getAgentViewFrequency() == 1.0 ) // VA_NORMAL AND VQ_HIGH (default)
  {
    if( iTriCounter % 3 == 0 )             // see will arrive first half cycle
    {
      iTimeSignal = (int)(iSimStep * PS->getFractionWaitSeeBegin() );
      iTriCounter = 0;
    }
    else if( iTriCounter % 3 == 1 )        // see will arrive 2nd half of cycle
    {
      iTimeSignal = (int)(iSimStep * PS->getFractionWaitSeeEnd() );
    }
    else                                   // no see will arrive
      iTimeSignal = (int)(iSimStep * PS->getFractionWaitNoSee( ) );
  }
  else if( WM->getAgentViewFrequency() == 2.0 ) // VA_WIDE AND VQ_HIGH
  {
    if( iTriCounter % 3 == 0 )              // see will arrive
    {
      iTimeSignal = (int)(iSimStep * PS->getFractionWaitSeeEnd() );
      iTriCounter = 0;
    }
    else                                   // no see will arrive
      iTimeSignal = (int)(iSimStep * PS->getFractionWaitNoSee() );
  }
  else                                     // VA_NARROW AND VQ_HIGH
    iTimeSignal = (int)(iSimStep * PS->getFractionWaitSeeEnd() );

  iTriCounter++;
  itv.it_value.tv_usec = iTimeSignal;
  setitimer( ITIMER_REAL, &itv, NULL );
}

/*! This method analyzes the type of the incoming message and calls the
    message that corresponds to this message.
    \param strMsg message that should be parsed.
    \return bool indicating whether the message was parsed or not */
bool SenseHandler::analyzeMessage( char *strMsg )
{
  switch( strMsg[1] )
  {
    case 'c':
        return analyzeChangePlayerTypeMessage( strMsg );      // ( c hange_
    case 'o':                                                 // ( o k
        if( strlen(strMsg) > 14 && strMsg[4] == 'c' && strMsg[10] == 'b' )
          analyzeCheckBall( strMsg );                         // (ok check_ball
        return true;
    case 's':
    {
      switch( strMsg[3] )
      {
        case 'e':
             if( strMsg[5] == 'g' )
                  return analyzeSeeGlobalMessage  ( strMsg ); // (se e_g
             else
                  return analyzeSeeMessage        ( strMsg ); // (se e
        case 'n': return analyzeSenseMessage      ( strMsg ); // (se n se
        case 'r': return analyzeServerParamMessage( strMsg ); // (se r ver_param
        default : break;
      }
    }
    case 'i':     return analyzeInitMessage       ( strMsg ); // ( i nit
    case 'h':     return analyzeHearMessage       ( strMsg ); // ( h ear
    case 'p':     return ( strMsg[8] == 't')
                  ? analyzePlayerTypeMessage ( strMsg )  // (player_ t ype
                  : analyzePlayerParamMessage( strMsg ); // (player_ p aram
    case 'e':     printf("%s", strMsg);                  // ( error
    default:      cerr << "(SenseHandler::analyzeMessage) " <<
                     "ignored message: " << strMsg << endl;
                  return false;
  }
  return false;
}


/*! This method analyzes a see Message. All information from the different
    objects that are stored in a see message is updated in the worldmodel.
    A see message looks like(see 0 ((g r) 64.1 13) ((f r t) 65.4 -16) ....
    \param strMsg message that should be parsed
    \return bool indicating whether the message was parsed correctly. */
bool SenseHandler::analyzeSeeMessage( char *strMsg )
{
  strcpy( WM->strLastSeeMessage, strMsg );
  Log.logWithTime( 2, " incoming see message" );

  ObjectT o;
  double  dDist, dDistChange,    dDirChange,     dTemp;
  int     iDir,  iBodyFacingDir, iHeadFacingDir, iTime = UnknownTime;
  Time    time = WM->getTimeLastSenseMessage();
  bool    isGoalie;

  iTime = Parse::parseFirstInt( &strMsg );         // get the time
  if( time.getTime() != iTime )
  {
    cerr << "(SenseHandler:analyzeSeeMessage) see with different time as sense:"
         << time.getTime() << " vs. " << iTime << endl;
    return false;
  }

  // reset the send pattern when previous cycle(s) no see arrived
  if( WM->getAgentViewFrequency() == 1.0 && // VA_NORMAL; previous cycle no see
      time.getTimeDifference( WM->getTimeLastSeeMessage() )== 2 )
    iTriCounter = 1;                // see will arive in 2nd half in next cycle
  else if( WM->getAgentViewFrequency() == 2.0 && // VA_WIDE; two cycles no see
           time.getTimeDifference( WM->getTimeLastSeeMessage() ) == 3 )
    iTriCounter = 1;                // no see will arrive next two cycles

  while( *strMsg != ')' )                          // " ((objname.." or ")"
  {
    dDist = dDistChange    = dDirChange     = dTemp = UnknownDoubleValue;
    iDir  = iBodyFacingDir = iHeadFacingDir =         UnknownIntValue;
    strMsg += 2;          // get the object name

    // get the object name from the first part of the string
    o = SoccerTypes::getObjectFromStr( &strMsg, &isGoalie, WM->getTeamName() );
    if( o == OBJECT_ILLEGAL )
    {
      Log.log( 4, "Illegal object in: ", WM->strLastSeeMessage );
      Log.log( 4, "rest of message: ", strMsg );
    }

    dTemp = Parse::parseFirstDouble( &strMsg );   // parse first value
    if ( *strMsg == ')' )                         // if it was the only value
      iDir = (int)dTemp;                          // it was the direction
    else
    {
      dDist = dTemp;                              // else it was distance
      iDir = Parse::parseFirstInt( &strMsg );     // and have to get direction
      if( *strMsg != ')' )                        // stil not finished
      {                                           // get dist and dir change
        dDistChange = Parse::parseFirstDouble( &strMsg );
        dDirChange  = Parse::parseFirstDouble( &strMsg );
      }
      if( *strMsg != ')' )                        // and stil not finished
      {                                           // get also facing dir.
        iBodyFacingDir = Parse::parseFirstInt( &strMsg );
        iHeadFacingDir = Parse::parseFirstInt( &strMsg );
      }
    }
    // skip ending bracket of object information.
    strMsg++;

    // process the parsed information (unread values are Unknown...)
    WM->processNewObjectInfo( o, time, dDist, iDir, dDistChange,
           dDirChange, (AngDeg)iBodyFacingDir, (AngDeg)iHeadFacingDir,
           isGoalie );
    }
  WM->mapUnknownPlayers( time );       // map players with unknown nr
  WM->setTimeLastSeeMessage( time );   // set time of last see message
                                       // this must be done after mapping since
                                       // otherwise mutex in WM is unlocked!
  return true;
}

/*! This method analyzes a See Message. All information from the different
    objects that is stored in a see message is updated in the worldmodel.
    A see message looks like(see 0 ((g r) 64.1 13) ((f r t) 65.4 -16) ....
    \param strMsg message that should be parsed
    \return bool indicating whether the message was parsed correctly. */
bool SenseHandler::analyzeSeeGlobalMessage( char *strMsg )
{
  Log.logWithTime( 2, " incoming see global message" );
  strcpy( WM->strLastSeeMessage, strMsg );

  ObjectT o;
  bool    isGoalie;
  double  dX, dY, dVelX, dVelY;
  int     iTime;
  AngDeg  angBody, angNeck;
  Time    time = WM->getCurrentTime();

  iTime = Parse::parseFirstInt( &strMsg );         // get the time
  time.updateTime( iTime );

  while( *strMsg != ')' )                          // " ((objname.." or ")"
  {
    dX = dY = dVelX = dVelY = UnknownDoubleValue;
    angBody = angNeck = UnknownAngleValue;
    strMsg += 2;          // go the start of the object name

    // get the object type at the current position in the string
    o = SoccerTypes::getObjectFromStr( &strMsg, &isGoalie, WM->getTeamName() );
    if( o == OBJECT_ILLEGAL )
    {
      Log.log( 4, "Illegal object" );
      Log.log( 4, "total messages: %s", WM->strLastSeeMessage );
      Log.log( 4, "rest of message: %s", strMsg );
    }

    dX = Parse::parseFirstDouble( &strMsg );        // parse first value
    dY = Parse::parseFirstDouble( &strMsg );        // parse second value
    if ( *strMsg != ')' )                           // if it was no flag
    {
      dVelX = Parse::parseFirstDouble( &strMsg );   // parse delta x
      dVelY = Parse::parseFirstDouble( &strMsg );   // parse delta y
      if( *strMsg != ')' )                          // stil not finished
      {                                             // get body and neck angle
        angBody = Parse::parseFirstDouble( &strMsg );
        angNeck = Parse::parseFirstDouble( &strMsg );
      }
    }
    // skip ending bracket of object information.
    strMsg++;

    // process the parsed information (unread values are Unknown...)
    WM->processSeeGlobalInfo( o, time, VecPosition(dX,dY),
                    VecPosition(dVelX,dVelY), angBody, angNeck );
    }
  WM->setTimeLastSeeGlobalMessage( time );  // set time last see global message
  return true;
}

/*! This method analyzes a sense message. All information from the player is
    parsed and updated in the WorldModel.
    A sense message looks like (sense_body 0 (view_mode high normal)
    (stamina 2000 1) (speed 0 0) (head_angle 0) (kick 0) (dash 0)
    (turn 0) (say 0) (turn_neck 0) (catch 0) (move 0) (change_view 0))
    \param strMsg message that should be parsed
    \return bool indicating whether the message was parsed correctly. */
bool SenseHandler::analyzeSenseMessage( char *strMsg )
{

  setTimeSignal();                            // set signal when to send action
  strcpy( WM->strLastSenseMessage, strMsg );

  int iTime = Parse::parseFirstInt( &strMsg );// get time
  Time timeOld = WM->getCurrentTime();
  Time timeNew = timeOld;
  timeNew.updateTime( iTime );

  if( timeNew.getTimeDifference( timeOld ) > 1 )
    Log.log( 1, "Missed a sense!!" );

  Log.logWithTime ( 2, "\n\nSENSE (%d)", timeNew.getTime() );
  Log.restartTimer( );
  Log.logWithTime ( 2, " alarm after %d", iTimeSignal );

  strMsg++;                                      // go to ( before view_mode

  Parse::gotoFirstOccurenceOf( ' ', &strMsg );   // skip view_mode
  strMsg++;                                      // skip space

  ViewQualityT vq = SoccerTypes::getViewQualityFromStr( strMsg ); // get quality
  Parse::gotoFirstOccurenceOf( ' ', &strMsg );
  strMsg++;                                      // skip space; get view_angle
  ViewAngleT va = SoccerTypes::getViewAngleFromStr( strMsg );
  double dStamina = Parse::parseFirstDouble( &strMsg );  // get stamina
  double dEffort  = Parse::parseFirstDouble( &strMsg );  // get effort

  double dSpeed   = Parse::parseFirstDouble( &strMsg );  // get speed
  AngDeg angSpeed = Parse::parseFirstDouble( &strMsg );  // get speed ang

  // minus sign since we store angle between neck and body and not vice versa
  int iHeadAngle = - Parse::parseFirstInt( &strMsg );    // get head_angle

  WM->processNewAgentInfo( vq, va, dStamina, dEffort, dSpeed,
        (AngDeg) angSpeed, (AngDeg)iHeadAngle );

  // set all number of performed commands
  WM->setNrOfCommands( CMD_KICK       , Parse::parseFirstInt( &strMsg ) );
  WM->setNrOfCommands( CMD_DASH       , Parse::parseFirstInt( &strMsg ) );
  WM->setNrOfCommands( CMD_TURN       , Parse::parseFirstInt( &strMsg ) );
  WM->setNrOfCommands( CMD_SAY        , Parse::parseFirstInt( &strMsg ) );
  WM->setNrOfCommands( CMD_TURNNECK   , Parse::parseFirstInt( &strMsg ) );
  WM->setNrOfCommands( CMD_CATCH      , Parse::parseFirstInt( &strMsg ) );
  WM->setNrOfCommands( CMD_MOVE       , Parse::parseFirstInt( &strMsg ) );
  WM->setNrOfCommands( CMD_CHANGEVIEW , Parse::parseFirstInt( &strMsg ) );

  WM->setTimeLastSenseMessage( timeNew );                     // set the time
  Log.logWithTime( 2, " end analyzing sense" );
  return true;
}

/*! This method analyzes an init message. All information from the
    initialization is parsed and updated in the WorldModel.
    An init message looks like (init [l|r] 10 before_kick_off)
    \param strMsg message that should be parsed
    \return bool indicating whether the message was parsed correctly. */
bool SenseHandler::analyzeInitMessage( char *strMsg )
{
  strMsg += 6;                                            // go to Side
  WM->setSide( SoccerTypes::getSideFromStr( strMsg ) );   // get and set Side
  int nr = Parse::parseFirstInt( &strMsg );               // get and set number
  if( nr == 0 )                                           // coach
  {
     WM->setPlayerNumber( nr );
     return true;
  }
  WM->setAgentObjectType( SoccerTypes::getTeammateObjectFromIndex( nr - 1 ) );
  WM->setPlayerNumber( nr );
  strMsg++;                                               // skip space to pm
  WM->setPlayMode( SoccerTypes::getPlayModeFromStr( strMsg ) ); // get playmode
  return true;
}

/*! This method analyzes a hear message. When the message is from the referee
    the message is parsed and the new play mode is set or the goal difference is
    adjusted. When the message comes from another player the method
    analyzePlayerMessage is called
    A hear message looks like (hear 0 self|referee|dir message)
    \param strMsg message that should be parsed
    \return bool indicating whether the message was parsed correctly. */
bool SenseHandler::analyzeHearMessage( char *strMsg )
{
  RefereeMessageT rm;
  PlayModeT       pm;
  strcpy( WM->strLastHearMessage, strMsg);

  Parse::parseFirstInt( &strMsg );                          // ignore time
  Time time = WM->getCurrentTime();

  switch( Parse::gotoFirstNonSpace( &strMsg ) )
  {
    case 'r':                                               // referee
      WM->setTimeLastRefereeMessage( time );
      Parse::gotoFirstOccurenceOf( ' ', &strMsg );          // go to start
      Parse::gotoFirstNonSpace   ( &strMsg      );          // and first part
      rm = SoccerTypes::getRefereeMessageFromStr( strMsg ); // get the ref msg
      Log.logWithTime( 2, " referee message: %s %s",
      SoccerTypes::getRefereeMessageStr(rm), WM->strLastHearMessage);
      pm = SoccerTypes::getPlayModeFromRefereeMessage( rm );// get play mode
      if( pm != PM_ILLEGAL )                                // from ref msg
        WM->setPlayMode( pm );                              // if was pm, set it

      if( rm == REFC_GOAL_LEFT )                            // goal left
      {
        if( WM->getSide() == SIDE_LEFT )
          WM->addOneToGoalDiff();
        else
          WM->subtractOneFromGoalDiff();
        WM->processSeeGlobalInfo( OBJECT_BALL, time, VecPosition( 0, 0 ),
             VecPosition( 0, 0 ), 0, 0 );
      }
      else if( rm == REFC_GOAL_RIGHT )                      // goal right
      {
        if( WM->getSide() == SIDE_RIGHT )
          WM->addOneToGoalDiff();
        else
          WM->subtractOneFromGoalDiff();
        WM->processSeeGlobalInfo( OBJECT_BALL, time, VecPosition( 0, 0 ),
             VecPosition( 0, 0 ), 0, 0 );        
      }
      else if( rm == REFC_GOALIE_CATCH_BALL_LEFT ||         // catch ball
               rm == REFC_GOALIE_CATCH_BALL_RIGHT )
        WM->processCatchedBall( rm, time );
      break;
    case 's':                                               // self
      break;                                                // do nothing
    default:                                                // from direction
        analyzePlayerMessage( strMsg );                     // from player
      break;
  }

  return true;
}

/*! This method analyzes a player message. Messages from teammates are encoded,
    so first the message is decoded to see if the message is from a teammate.
    When this is the case all information (position and velocity from sender
    and ball and position from all opponents and teammates) are parsed and
    passed to the WorldModel. The format of the received message is:
    (say ***** my_x my_y my_vel_x my_vel_y
             (ball_x ball_y ball_vel_x ball_vel_y ball_conf)
             (team_nr pos_x pos_y conf ...)
             (opp_nr pos_x pos_y  conf ...))
    \param strMsg string that contains the player message
    \return bool indicating whether the message was succesfully parsed. */
bool SenseHandler::analyzePlayerMessage( char *strMsg )
{
  Parse::parseFirstInt( &strMsg );        // skip direction
  Parse::gotoFirstNonSpace( &strMsg );    // skip space
  strMsg++;                               // skip " (=quote)

  // first character is team side
  char cSide = (WM->getSide() == SIDE_LEFT) ? 'l' : 'r' ;

  if( cSide != strMsg[0] )                // not team side probably msg from opp
    return false;

  strMsg++;                              // go to next

  if( strlen(strMsg) < 7    ||           // encoding always at least 7 chars
      strMsg[0]      < 'a'  ||           // and between 'a' and 'a' + 9 since
      strMsg[0]      > 'a'+ 9 )          // it is a number added to 'a'.
    return false;                        // not right -> message from opponent

  int i[4];                              // decode message by subracting
  i[0] =        (int)(strMsg[0] - 'a');  // string 'acegi' from message.
  i[1] =        (int)(strMsg[1] - 'c');  // resulting number indicate cycle and
  i[2] =        (int)(strMsg[2] - 'e');  // player number.
  i[3] =        (int)(strMsg[3] - 'g');
  int iPlayer = (int)(strMsg[4] - 'i');

  int iTime = i[0]*1000+i[1]*100+i[2]*10+i[3];

  int    iDiff = iTime - WM->getCurrentCycle()%10000;
  double dDiff = iDiff/100;

  if( fabs(iDiff) > 2 && strMsg[0] != 'F' )
  {
    printf( "(SenseHandler::%s) information is outdated:%d %d in:",
                      __FUNCTION__, iTime, WM->getCurrentCycle() );
    printf( "%s\n", WM->strLastHearMessage );
    return false;
  }

  // is message from teammate, so start retrieving information
  VecPosition pos, vel;
  double      dConf;
  ObjectT     obj;

  // parse information from player that sent the message
  pos.setX( Parse::parseFirstDouble( &strMsg ) );
  pos.setY( Parse::parseFirstDouble( &strMsg ) );
  vel.setX( Parse::parseFirstDouble( &strMsg ) );
  vel.setY( Parse::parseFirstDouble( &strMsg ) );
  obj = SoccerTypes::getTeammateObjectFromIndex( iPlayer - 1 );

  WM->processPerfectHearInfo( obj, pos, 1.00 ); // vel is not parsed.

  // parse the information from the ball, the confidence is subtracted with
  // dDiff that denotes the confidence difference between the sent time and
  // the heard time. When message is heard one cycle later than it is sent,
  // confidence should be lower.
  Parse::gotoFirstOccurenceOf( '(', &strMsg );
  strMsg++;
  if( *strMsg != ')' )                                  // ball info available
  {
    pos.setX( Parse::parseFirstDouble( &strMsg ) );
    pos.setY( Parse::parseFirstDouble( &strMsg ) );
    vel.setX( Parse::parseFirstDouble( &strMsg ) );
    vel.setY( Parse::parseFirstDouble( &strMsg ) );
    dConf = Parse::parseFirstDouble( &strMsg );
    WM->processPerfectHearInfoBall( pos, vel, dConf - dDiff );
  }

  Parse::gotoFirstOccurenceOf( ')', &strMsg );          // go to teammate info
  strMsg++; // skip ')'
  strMsg++; // skip '('
  while( *strMsg != ')' )                               // as long as new info
  {
    iPlayer = Parse::parseFirstInt( &strMsg );          // get the number
    if( iPlayer == -1 )                                 // if illegal
      obj = OBJECT_TEAMMATE_UNKNOWN;                    // we don't know nr
    else
      obj = SoccerTypes::getTeammateObjectFromIndex( iPlayer - 1 );
    pos.setX( Parse::parseFirstDouble( &strMsg ) );
    pos.setY( Parse::parseFirstDouble( &strMsg ) );
    dConf = Parse::parseFirstDouble( &strMsg ) ;
    Parse::gotoFirstNonSpace( &strMsg );
    WM->processPerfectHearInfo( obj, pos, dConf - dDiff ); // process all info
  }

  bool bIsGoalie = false;

  strMsg++; // skip ')'                                 // go to opponent info
  strMsg++; // skip '('
  while( *strMsg != ')' )
  {
    iPlayer = Parse::parseFirstInt( &strMsg );
    if( iPlayer == -1 )                                 // if illegal
      obj = OBJECT_OPPONENT_UNKNOWN;                    // we don't know nr
    else
      obj = SoccerTypes::getOpponentObjectFromIndex( iPlayer - 1 );
    bIsGoalie = ( *strMsg == 'g' );                     // i.e. 1g
    pos.setX( Parse::parseFirstDouble( &strMsg ) );
    pos.setY( Parse::parseFirstDouble( &strMsg ) );
    dConf = Parse::parseFirstDouble( &strMsg ) ;
    Parse::gotoFirstNonSpace( &strMsg );
    WM->processPerfectHearInfo( obj, pos, dConf - dDiff, bIsGoalie );
  }
  return true;
}

/*! This method analyzes the check_ball message that is only received by the
    coach. It sets the information in the Worldmodel what the status of the
    ball is.
    The format is as follows (check_ball <time> <status>).
    \param strMsg string that contains the check_ball message
    \return bool indicating whether update succeeded. */
bool SenseHandler::analyzeCheckBall( char *strMsg )
{
  WM->setTimeCheckBall( Parse::parseFirstInt( &strMsg ) );
  strMsg++;
  WM->setCheckBallStatus( SoccerTypes::getBallStatusFromStr( strMsg ) );
  return true;
}

/*! This method analyzes the change player type message. This method checks
    whether the player that changed type equals the agent. When this is the
    case, it adjust the ServerSettings according to the values associated
    with this player type.
    \param strMsg string that contains the player type message.
    \return bool indicating whether player type of agent changed. */
bool SenseHandler::analyzeChangePlayerTypeMessage( char *strMsg )
{
  int iPlayer = Parse::parseFirstInt( &strMsg );
  if( *strMsg != ')' && WM->getPlayerNumber() == iPlayer )
  {
    int iPlayerType = Parse::parseFirstInt( &strMsg );
    return WM->updateSSToHeteroPlayerType( iPlayerType );
  }
  return false;
}

/*! This method analyzes the server_param message. This message contains all
    the server parameters. All settings of the ServerSettings are changed
    according to the supplied values. This makes the reading from a server
    configuration file obsolete.
    \param strMsg string message with all the server parameters
    \return booli indicating whether string was parsed. */
bool SenseHandler::analyzeServerParamMessage( char *strMsg )
{
  // get the server parameters one by one
  SS->setGoalWidth            ( Parse::parseFirstDouble( &strMsg ) );
  SS->setInertiaMoment        ( Parse::parseFirstDouble( &strMsg ) );
  SS->setPlayerSize           ( Parse::parseFirstDouble( &strMsg ) );
  SS->setPlayerDecay          ( Parse::parseFirstDouble( &strMsg ) );
  SS->setPlayerRand           ( Parse::parseFirstDouble( &strMsg ) );
  SS->setPlayerWeight         ( Parse::parseFirstDouble( &strMsg ) );
  SS->setPlayerSpeedMax       ( Parse::parseFirstDouble( &strMsg ) );
  SS->setPlayerAccelMax       ( Parse::parseFirstDouble( &strMsg ) );
  SS->setStaminaMax           ( Parse::parseFirstDouble( &strMsg ) );
  SS->setStaminaIncMax        ( Parse::parseFirstDouble( &strMsg ) );
  Parse::parseFirstDouble( &strMsg ); // skip recover_init = 0
  SS->setRecoverDecThr        ( Parse::parseFirstDouble( &strMsg ) );
  SS->setRecoverMin           ( Parse::parseFirstDouble( &strMsg ) );
  SS->setRecoverDec           ( Parse::parseFirstDouble( &strMsg ) );
  Parse::parseFirstDouble( &strMsg ); // skip effort_init = 0
  SS->setEffortDecThr         ( Parse::parseFirstDouble( &strMsg ) );
  SS->setEffortMin            ( Parse::parseFirstDouble( &strMsg ) );
  SS->setEffortDec            ( Parse::parseFirstDouble( &strMsg ) );
  SS->setEffortIncThr         ( Parse::parseFirstDouble( &strMsg ) );
  SS->setEffortInc            ( Parse::parseFirstDouble( &strMsg ) );
  SS->setKickRand             ( Parse::parseFirstDouble( &strMsg ) );
  Parse::parseFirstDouble( &strMsg ); // skip team_actuator noise
  Parse::parseFirstDouble( &strMsg ); // skip prand_factor_l
  Parse::parseFirstDouble( &strMsg ); // skip prand_factor_r
  Parse::parseFirstDouble( &strMsg ); // skip kick_rand_factor_l
  Parse::parseFirstDouble( &strMsg ); // skip kick_rand_factor_r
  SS->setBallSize             ( Parse::parseFirstDouble( &strMsg ) );
  SS->setBallDecay            ( Parse::parseFirstDouble( &strMsg ) );
  SS->setBallRand             ( Parse::parseFirstDouble( &strMsg ) );
  SS->setBallWeight           ( Parse::parseFirstDouble( &strMsg ) );
  SS->setBallSpeedMax         ( Parse::parseFirstDouble( &strMsg ) );
  SS->setBallAccelMax         ( Parse::parseFirstDouble( &strMsg ) );
  SS->setDashPowerRate        ( Parse::parseFirstDouble( &strMsg ) );
  SS->setKickPowerRate        ( Parse::parseFirstDouble( &strMsg ) );
  SS->setKickableMargin       ( Parse::parseFirstDouble( &strMsg ) );
  Parse::parseFirstDouble( &strMsg ); // skip control radius
  Parse::parseFirstDouble( &strMsg ); // skip control radius width
  SS->setMaxPower             ( Parse::parseFirstInt( &strMsg ) );
  SS->setMinPower             ( Parse::parseFirstInt( &strMsg ) );
  SS->setMaxMoment            ( Parse::parseFirstInt( &strMsg ) );
  SS->setMinMoment            ( Parse::parseFirstInt( &strMsg ) );
  SS->setMaxNeckMoment        ( Parse::parseFirstInt( &strMsg ) );
  SS->setMinNeckMoment        ( Parse::parseFirstInt( &strMsg ) );
  SS->setMaxNeckAng           ( Parse::parseFirstInt( &strMsg ) );
  SS->setMinNeckAng           ( Parse::parseFirstInt( &strMsg ) );
  SS->setVisibleAngle         ( Parse::parseFirstDouble( &strMsg ) );
  SS->setVisibleDistance      ( Parse::parseFirstDouble( &strMsg ) );
  SS->setWindDir              ( Parse::parseFirstDouble( &strMsg ) );
  SS->setWindForce            ( Parse::parseFirstDouble( &strMsg ) );
  Parse::parseFirstDouble( &strMsg ); // skip winang
  SS->setWindRand             ( Parse::parseFirstDouble( &strMsg ) );
  SS->setMaximalKickDist      ( Parse::parseFirstDouble( &strMsg ) );
  SS->setCatchableAreaL       ( Parse::parseFirstDouble( &strMsg ) );
  SS->setCatchableAreaW       ( Parse::parseFirstDouble( &strMsg ) );
  SS->setCatchProbability     ( Parse::parseFirstDouble( &strMsg ) );
  SS->setGoalieMaxMoves       ( Parse::parseFirstInt( &strMsg ) );
  SS->setCkickMargin          ( Parse::parseFirstDouble( &strMsg ) );
  SS->setOffsideActiveAreaSize( Parse::parseFirstDouble( &strMsg ) );
  Parse::parseFirstDouble( &strMsg ); // skip win_no: wind factor is none
  Parse::parseFirstDouble( &strMsg ); // skip win_random: wind factor is random
  SS->setSayCoachCntMax       ( Parse::parseFirstInt( &strMsg ) );
  SS->setSayCoachMsgSize      ( Parse::parseFirstInt( &strMsg ) );
  SS->setClangWinSize         ( Parse::parseFirstInt( &strMsg ) );
  SS->setClangDefineWin       ( Parse::parseFirstInt( &strMsg ) );
  SS->setClangMetaWin         ( Parse::parseFirstInt( &strMsg ) );
  SS->setClangAdviceWin       ( Parse::parseFirstInt( &strMsg ) );
  SS->setClangInfoWin         ( Parse::parseFirstInt( &strMsg ) );
  SS->setClangMessDelay       ( Parse::parseFirstInt( &strMsg ) );
  SS->setClangMessPerCycle    ( Parse::parseFirstInt( &strMsg ) );
  SS->setHalfTime             ( Parse::parseFirstInt( &strMsg ) );
  SS->setSimulatorStep        ( Parse::parseFirstInt( &strMsg ) );
  SS->setSendStep             ( Parse::parseFirstInt( &strMsg ) );
  SS->setRecvStep             ( Parse::parseFirstInt( &strMsg ) );
  SS->setSenseBodyStep        ( Parse::parseFirstInt( &strMsg ) );
  Parse::parseFirstInt( &strMsg ); // ignore lcm (lowest common multiplier?)
  SS->setSayMsgSize           ( Parse::parseFirstInt( &strMsg ) );
  SS->setHearMax              ( Parse::parseFirstInt( &strMsg ) );
  SS->setHearInc              ( Parse::parseFirstInt( &strMsg ) );
  SS->setHearDecay            ( Parse::parseFirstInt( &strMsg ) );
  SS->setCatchBanCycle        ( Parse::parseFirstInt( &strMsg ) );
  SS->setSlowDownFactor       ( Parse::parseFirstInt( &strMsg ) );
  SS->setUseOffside           ( Parse::parseFirstInt( &strMsg ) );
  SS->setForbidKickOffOffside ( Parse::parseFirstInt( &strMsg ) );
  SS->setOffsideKickMargin    ( Parse::parseFirstDouble( &strMsg ) );
  SS->setAudioCutDist         ( Parse::parseFirstDouble( &strMsg ) );
  SS->setQuantizeStep         ( Parse::parseFirstDouble( &strMsg ) );
  SS->setQuantizeStepL        ( Parse::parseFirstDouble( &strMsg ) );
  Parse::parseFirstDouble( &strMsg ); // skip dir_qstep
  Parse::parseFirstDouble( &strMsg ); // skip dist_qstep_l
  Parse::parseFirstDouble( &strMsg ); // skip dist_qstep_r
  Parse::parseFirstDouble( &strMsg ); // skip land_qstep_l
  Parse::parseFirstDouble( &strMsg ); // skip land_qstep_r
  Parse::parseFirstDouble( &strMsg ); // skip dir_qstep_l
  Parse::parseFirstDouble( &strMsg ); // skip dir_qstep_r
  Parse::parseFirstInt( &strMsg );    // skip CoachMode (coach allowed or not)
  Parse::parseFirstInt( &strMsg );    // skip CwRMode (referee allowed with c)
  Parse::parseFirstInt( &strMsg );    // old_hear (old format for coach hear)
  Parse::parseFirstInt( &strMsg );    // setCoachLookInterval equals viStep?
  Parse::parseFirstInt( &strMsg );    // skip start_goal_l
  Parse::parseFirstInt( &strMsg );    // skip start_goal_r
  Parse::parseFirstInt( &strMsg );    // skip fullstate_l
  Parse::parseFirstInt( &strMsg );    // skip fullstate_r
  Parse::parseFirstInt( &strMsg );    // skip drop_time

  return true;
}

/*! This method analyze a player type message. This message contains the
    values associated with a specific heterogeneous player type. The values
    are parsed from the message and supplied to the WorldModel method
    processNewHeteroPlayer.
    \param strMsg string that contains the player type information
    \return bool indicating whether the message was parsed correctly. */
bool SenseHandler::analyzePlayerTypeMessage ( char *strMsg )
{
  // analyze all heterogeneous player information
  int    iIndex           = Parse::parseFirstInt( &strMsg );
  double dPlayerSpeedMax  = Parse::parseFirstDouble( &strMsg );
  double dStaminaIncMax   = Parse::parseFirstDouble( &strMsg );
  double dPlayerDecay     = Parse::parseFirstDouble( &strMsg );
  double dInertiaMoment   = Parse::parseFirstDouble( &strMsg );
  double dDashPowerRate   = Parse::parseFirstDouble( &strMsg );
  double dPlayerSize      = Parse::parseFirstDouble( &strMsg );
  double dKickableMargin  = Parse::parseFirstDouble( &strMsg );
  double dKickRand        = Parse::parseFirstDouble( &strMsg );
  double dExtraStamina    = Parse::parseFirstDouble( &strMsg );
  double dEffortMax       = Parse::parseFirstDouble( &strMsg );
  double dEffortMin       = Parse::parseFirstDouble( &strMsg );
  WM->processNewHeteroPlayer( iIndex, dPlayerSpeedMax, dStaminaIncMax,
      dPlayerDecay,    dInertiaMoment, dDashPowerRate, dPlayerSize,
      dKickableMargin, dKickRand,      dExtraStamina,  dEffortMax,
      dEffortMin );
  return true;
}

/*! This method analyzes the player_param message that indicates the ranges of
    the possible values for the heterogeneous player types. Nothing is done
    with this information.
    \param strMsg string that contains the player_param message.
    \bool will always be true. */
bool SenseHandler::analyzePlayerParamMessage( char *strMsg )
{
  return true;
}

/******************************************************************************/
/********************** TESTING PURPOSES **************************************/
/******************************************************************************/

/*
int main( void )
{
  Connection c( "localhost", 6000 );
  WorldModel wm;
  SenseHandler i( &c,  &wm );
  i.analyzeMessage( "(see 0 ((g r) 64.1 13) ((f r t) 65.4 -16) ((f r b) 79 38) ((f p r t) 46.1 -6) ((f p r c) 48.4 18) ((f p r b) 58 37) ((f g r t) 62.8 7) ((f g r b) 66 19) ((f t r 20) 38.5 -38) ((f t r 30) 46.5 -30) ((f t r 40) 55.7 -25) ((f t r 50) 64.7 -21) ((f b r 50) 80.6 41) ((f r t 30) 69.4 -12) ((f r t 20) 67.4 -4) ((f r t 10) 67.4 4) ((f r 0) 69.4 12) ((f r b 10) 72.2 20) ((f r b 20) 75.9 27) ((f r b 30) 81.5 33) ((l r) 62.8 -89))" );
   cout << "2" << endl;
  i.analyzeMessage( "(see 0 ((g l) 49.9 -24) ((f l t) 50.9 14) ((f p l t) 31.5 1 0 0) ((f p l c) 34.5 -33) ((f g l t) 47.9 -17) ((f g l b) 52.5 -32) ((f t l 50) 50.9 20) ((f t l 40) 42.5 26) ((f t l 30) 34.8 36) ((f l t 30) 54.6 8) ((f l t 20) 53 -2) ((f l t 10) 53 -12) ((f l 0) 54.6 -23) ((f l b 10) 58 -32) ((f l b 20) 62.8 -41) ((p \"l\" 2) 5 -7 0 0 172 172) ((l l) 47.9 82))" );
  c.disconnect();
  cout << "exit" << endl ;

}

*/
